# Generated by Django 2.0.5 on 2018-05-18 11:47

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('genomix_workflows', '0002_auto_20180516_1649'),
    ]

    operations = [
        migrations.CreateModel(
            name='Reference',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('object_id', models.PositiveIntegerField(db_index=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('workflow', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='references', to='genomix_workflows.Workflow')),
            ],
            options={
                'verbose_name': 'Reference',
                'verbose_name_plural': 'References',
            },
        ),
        migrations.DeleteModel(
            name='Connection',
        ),
        migrations.AddField(
            model_name='workflowinstance',
            name='state',
            field=models.PositiveSmallIntegerField(choices=[(1, 'Active'), (2, 'Complete'), (3, 'Hold'), (4, 'Canceled')], default=1),
        ),
        migrations.AlterField(
            model_name='workflowinstance',
            name='parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='genomix_workflows.WorkflowInstance'),
        ),
        migrations.AlterUniqueTogether(
            name='reference',
            unique_together={('content_type', 'object_id', 'workflow')},
        ),
    ]



# -*- coding: utf-8
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver

from . import models, choices


@receiver(post_save, sender=models.WorkflowInstance)
def workflow_instance_post_save(sender, instance, created, **kwargs):
    """
    After WorkflowInstance.save is called, we look for all the related tasks in
    the workflow and make a copy of them in the TaskInstance.
    """

    pending = getattr(choices.STATUS, 'PENDING')

    if created:
        # Create an Instance of each Task using the referenced workflow
        for task in instance.workflow.task.all():
            models.TaskInstance.objects.create(task=task,
                                               workflow_instance=instance,
                                               status=None)

        if instance.parent is None:
            first_task_id = next(iter(instance.topological_sort() or []), None)
            if first_task_id:
                task = models.TaskInstance.objects.get(pk=first_task_id)
                if task.status is None:
                        task.status = pending
                        task.save()


@receiver(pre_save, sender=models.TaskInstance)
def task_instance_pre_save(sender, instance, **kwargs):
    """
    After the task is completed, it should set the status of the next
    task instances to PENDING.
    If the task instance was the last one in the workflow, it should set
    the workflow state to COMPLETE.
    """

    complete = getattr(choices.STATUS, 'COMPLETE')
    pending = getattr(choices.STATUS, 'PENDING')
    running = getattr(choices.STATUS, 'RUNNING')
    workflow_active = getattr(choices.WORKFLOW_STATE, 'ACTIVE')
    workflow_complete = getattr(choices.WORKFLOW_STATE, 'COMPLETE')

    # Check if it's an update
    if instance.pk is not None:
        old_instance = models.TaskInstance.objects.get(pk=instance.pk)

        # Only allow change to complete/running when all previous tasks are complete
        if old_instance.status is None and instance.status in (complete, pending, running,):
            # Check if it can be completed
            if not instance.__class__.objects.combined_previous_is_complete(instance):
                raise Exception('Can only initiate a task if all previous tasks are completed')

        # If a task completion was rolled back, workflow status should change back to active
        if (instance.status != complete) and (old_instance.status == complete):
            # prevent roll back if next tasks are already complete
            related_next_instances = instance.__class__.objects.combined_next_task_instances(instance)
            for task in related_next_instances:
                if task.status in (complete, running,):
                    raise Exception('Cannot roll back a Task if any of the next tasks have already been initiated')
            for task in related_next_instances:
                if task.status == pending:
                    task.status = None
                    task.save()

            workflow = instance.workflow_instance
            if workflow.state == workflow_complete:
                workflow.state = workflow_active
                workflow.save()


@receiver(post_save, sender=models.TaskInstance)
def task_instance_post_save(sender, created, instance, **kwargs):
    """
    After the task is completed, it should set the status of the next
    task instances to PENDING.
    If the task instance was the last one in the workflow, it should set
    the workflow state to COMPLETE.
    """

    complete = getattr(choices.STATUS, 'COMPLETE')
    pending = getattr(choices.STATUS, 'PENDING')
    workflow_complete = getattr(choices.WORKFLOW_STATE, 'COMPLETE')

    workflow = instance.workflow_instance

    if not created:
        # Status: xxx -> Complete
        if (instance.status == complete):
            # check if all tasks are completed then change workflow state to complete
            if workflow.all_tasks_completed():
                workflow.state = workflow_complete
                workflow.save()

            combined_next_instances = instance.__class__.objects.combined_next_task_instances(instance)
            for task in combined_next_instances:
                if task.__class__.objects.combined_previous_is_complete(task) and task.status is None:
                    task.status = pending
                    task.save()



# Generated by Django 2.0.5 on 2018-06-04 16:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('genomix_workflows', '0004_auto_20180601_1216'),
    ]

    operations = [
        migrations.AddField(
            model_name='reference',
            name='workflow_category',
            field=models.CharField(default='order', max_length=100),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='taskinstance',
            name='status',
            field=models.PositiveSmallIntegerField(choices=[(1, 'Pending'), (2, 'Complete'), (3, 'Running'), (4, 'Hold'), (5, 'Skip'), (6, 'Failed')], default=1),
        ),
        migrations.AlterField(
            model_name='workflowinstance',
            name='state',
            field=models.PositiveSmallIntegerField(choices=[(1, 'Active'), (2, 'Complete'), (3, 'Hold'), (4, 'Cancelled'), (5, 'Failed')], default=1),
        ),
    ]



# -*- coding: utf-8
from django.contrib import admin
from django.utils.safestring import mark_safe
from reversion.admin import VersionAdmin
import json

from . import models


class NextTaskInline(admin.TabularInline):
    model = models.Task.previous.through
    fk_name = 'to_task'
    verbose_name = u"Next"
    verbose_name_plural = u"Nexts"


class TaskAdmin(admin.ModelAdmin):
    model = models.Task
    list_display = ('id', 'label', 'get_all_previous')
    inlines = [
        NextTaskInline
    ]
    filter_horizontal = ('previous',)


class WorkflowAdmin(admin.ModelAdmin):
    model = models.Workflow
    list_display = ('id', 'label', 'version')
    filter_horizontal = ('task',)


class WorkflowInstanceAdmin(VersionAdmin):
    model = models.WorkflowInstance
    list_display = ('id', 'workflow', 'content_type', 'object_id',)
    readonly_fields = (
        '_children',
        '_all_related',
        '_current_status',
        '_combined_current_status',
        '_graph_diagram',
        '_topological_sort',
        '_combined_topological_sort',
        '_combined_graph_diagram',
        '_graph_json',
        '_combined_graph_json',
    )
    raw_id_fields = ('parent',)

    class Media:
        css = {
            "all": ('css/graph.css',)
        }
        js = ('https://d3js.org/d3.v5.js',
              'js/dagre-d3.js',
              'js/graph.js',)

    def _children(self, obj):
        return obj.get_all_children()

    def _all_related(self, obj):
        return obj.all_related_workflow_instance_ids()

    def _current_status(self, obj):
        return obj.current_status()

    def _combined_current_status(self, obj):
        return self.model.objects.combined_current_status(obj)

    def _topological_sort(self, obj):
        return obj.topological_sort()

    def _combined_topological_sort(self, obj):
        return self.model.objects.combined_topological_sort(obj)

    def _graph_diagram(self, obj):
        json_string = "'" + json.dumps(obj.get_graph_json()) + "'"
        result = u'<svg class="graph1" width="700" height="400" data=%s ></svg>' % json_string
        return mark_safe(result)

    def _graph_json(self, obj):
        json_string = obj.get_graph_json()
        result = u'<pre>%s</pre>' % json_string
        return mark_safe(result)

    def _combined_graph_diagram(self, obj):
        json_string = "'" + json.dumps(self.model.objects.get_combined_graph_json(obj)) + "'"
        result = u'<svg class="graph2" width="700" height="500" data=%s ></svg>' % json_string
        return mark_safe(result)

    def _combined_graph_json(self, obj):
        json_string = self.model.objects.get_combined_graph_json(obj)
        result = u'<pre>%s</pre>' % json_string
        return mark_safe(result)


class TaskInstanceAdmin(VersionAdmin):
    model = models.TaskInstance
    list_display = ('id', 'task', 'workflow_instance', 'status', '_previous',)
    readonly_fields = ('_previous', '_combined_previous', '_next', '_combined_next')
    raw_id_fields = ('workflow_instance',)

    def _previous(self, obj):
        return "\n".join([str(ti) for ti in obj.previous_task_instances()])

    def _combined_previous(self, obj):
        return "\n".join([str(ti) for ti in self.model.objects.combined_previous_task_instances(obj)])

    def _next(self, obj):
        return "\n".join([str(ti) for ti in obj.next_task_instances()])

    def _combined_next(self, obj):
        return "\n".join([str(ti) for ti in self.model.objects.combined_next_task_instances(obj)])


class ReferenceAdmin(admin.ModelAdmin):
    model = models.Reference
    list_display = ('id', 'workflow_category', 'content_type', 'object_id', 'workflow')


admin.site.register(models.Reference, ReferenceAdmin)
admin.site.register(models.Task, TaskAdmin)
admin.site.register(models.TaskInstance, TaskInstanceAdmin)
admin.site.register(models.Workflow, WorkflowAdmin)
admin.site.register(models.WorkflowInstance, WorkflowInstanceAdmin)




# -*- coding: utf-8 -*-
from rest_framework import routers


app_name = 'genomix_workflows'

# Simple router that can be extended in a main API
router = routers.SimpleRouter()

# Router with API Root
default_router = routers.DefaultRouter()

urlpatterns = default_router.urls



# -*- coding: utf-8 -*-
from model_utils import Choices


STATUS = Choices(
    (1, 'PENDING', 'Pending'),
    (2, 'COMPLETE', 'Complete'),
    (3, 'RUNNING', 'Running'),
    (4, 'HOLD', 'Hold'),
    (5, 'SKIP', 'Skip'),
    (6, 'FAILED', 'Failed'),
)

WORKFLOW_STATE = Choices(
    (1, 'ACTIVE', 'Active'),
    (2, 'COMPLETE', 'Complete'),
    (3, 'HOLD', 'Hold'),
    (4, 'CANCELLED', 'Cancelled'),
    (5, 'FAILED', 'Failed'),
)



# -*- coding: utf-8 -*-
from django.db.models import Q, Manager
import json
import networkx as nx
from networkx import json_graph

from . import models, choices


class WorkflowManager(Manager):

    def combined_task_instances(self, obj):
        wf_instances = []

        combined_task_instances = models.TaskInstance.objects

        if obj.parent is None:
            wf_instances.extend(obj.get_all_children())
        else:
            wf_instances.extend(obj.parent.get_all_children())
        combined_task_instances = combined_task_instances.filter(workflow_instance__in=wf_instances)

        return combined_task_instances

    def combined_current_status(self, obj):

        running = getattr(choices.STATUS, 'RUNNING')
        pending = getattr(choices.STATUS, 'PENDING')
        workflow_complete = getattr(choices.WORKFLOW_STATE, 'COMPLETE')
        current_status = []
        combined_task_instances = obj.__class__.objects.combined_task_instances(obj)

        if obj.state != workflow_complete:
            current_status = combined_task_instances.filter(status__in=[pending, running])

        return current_status

    def get_combined_graph(self, obj):

        G = nx.DiGraph()
        combined_task_instances = obj.__class__.objects.combined_task_instances(obj)

        for task_instance in combined_task_instances:

            previous_tasks = task_instance.task.previous.all()

            for previous in previous_tasks:
                # find previous in task instances if exists
                # if (task_instance.task.id == previous.id):
                if combined_task_instances.filter(task__id=previous.id):
                    previous_task_instances = combined_task_instances.filter(task__id=previous.id)
                    # also check to make sure the two instance workflows are directly related
                    if task_instance.workflow_instance.parent is not None:
                        previous_task_instances = previous_task_instances.filter(
                            Q(workflow_instance=task_instance.workflow_instance.parent) |
                            Q(workflow_instance=task_instance.workflow_instance)
                        )
                    for pti in previous_task_instances:
                        # add the previous instance and its edge to graph
                        G.add_node(task_instance.id,
                                   status=task_instance.get_status_display(),
                                   label=task_instance.task.label,
                                   group=str(task_instance.workflow_instance),
                                   )
                        G.add_node(pti.id,
                                   status=pti.get_status_display(),
                                   label=pti.task.label,
                                   group=str(pti.workflow_instance),
                                   )
                        G.add_edge(pti.id, task_instance.id)
        return G

    def get_combined_graph_json(self, obj):
        G = obj.__class__.objects.get_combined_graph(obj)
        graph = json.dumps(json_graph.node_link_data(G), indent=2)
        return graph

    def combined_topological_sort(self, obj):
        G = obj.__class__.objects.get_combined_graph(obj)
        sorted = list(nx.topological_sort(G))
        return sorted


class TaskManager(Manager):

    def combined_previous_task_instances(self, obj):
        previous_instances = []
        related_wf_instance_ids = obj.workflow_instance.all_related_workflow_instance_ids()
        all_related_task_instances = models.TaskInstance.objects.filter(
            workflow_instance__id__in=related_wf_instance_ids
        )

        # Find all previous from workflows other than it's own:
        for ti in all_related_task_instances:
            # check if the task is one of previous ones:
            if ti.task in obj.task.previous.all():
                previous_i = all_related_task_instances.filter(id=ti.id).first()
                # add the previous instance to the list
                previous_instances.append(previous_i)
        # Now add previous task from it's own workflow too:
        previous_instances.extend(obj.previous_task_instances())
        return previous_instances

    def combined_next_task_instances(self, obj):
        next_instances = []
        related_wf_instance_ids = obj.workflow_instance.all_related_workflow_instance_ids()
        all_related_task_instances = models.TaskInstance.objects.filter(
            workflow_instance__id__in=related_wf_instance_ids
        )

        # Find all nexts from workflows other than it's own:
        for ti in all_related_task_instances:
            # check if the task is one of next ones:
            if ti.task in obj.task.next_tasks.all():
                next_i = all_related_task_instances.filter(id=ti.id).first()
                # add the next instance to the list
                next_instances.append(next_i)
        # Add next task from it's own workflow too:
        next_instances.extend(obj.next_task_instances())
        return next_instances

    def combined_previous_is_complete(self, obj):
        """
        Check combined previous task instances to see if they are all completed
        """
        result = True
        complete = getattr(choices.STATUS, 'COMPLETE')

        combined_previous = obj.__class__.objects.combined_previous_task_instances(obj)

        for task in combined_previous:
            if task.status != complete:
                result = False

        return result
